#define _GNU_SOURCE

#include <linux/netfilter.h>
#include <libipq.h>
#include <stdio.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <dirent.h>
#include <netdb.h>
#include <sys/un.h>
#include <signal.h>
#include <glib.h>
#include <grp.h>
#include <pwd.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/tcp.h>


#include "common.h"

#define BUFSIZE 1024*20
#define MAX_CLIENTS 20

typedef struct {
    int sock;
    pid_t pid;
    GSList *send_queue;
    guint send_index;
    guint8 *recv_buf;
    guint recv_index;
    message_t *recv_msg;
    gboolean dead;
} client_t;

GSList *clients = NULL;

GSList *packets = NULL;
guint n_packets;

volatile int signum = 0;

struct sockaddr_un sa;
socklen_t sa_len = 0;

fd_set listen_wfds;
fd_set listen_rfds;

void sigfunc(int s) {
    signum = s;
}

gint unix_socket(char *p) {
    struct sockaddr_un addr;
    int sock;
    mode_t save;
    
    save = umask(0007);
                                                                                                                                                
    if ((sock = socket(PF_LOCAL, SOCK_STREAM, 0)) < 0) {
        umask(save);
        perror("socket(PF_LOCAL, SOCK_STREAM, 0)");
        return -1;
    }

    addr.sun_family = AF_LOCAL;
    strncpy(addr.sun_path, p, sizeof(addr.sun_path));
    addr.sun_path[sizeof(addr.sun_path)-1] = 0;

    if (bind(sock, (struct sockaddr *) &addr, SUN_LEN(&addr)) < 0) {
        umask(save);
        perror("bind()");
        return -1;
    }

    umask(save);

    if (listen(sock, 1) < 0) {
        perror("listen()");
        return -1;
    }
    
    return sock;
}

void message_deref(message_t *m) {
    g_assert(m);
    m->ref--;
    if (m->ref <= 0)
        g_free(m);
}


gint client_send_work(client_t *c) {
    g_assert(c);
    if (c->send_queue) {
        size_t l;
        ssize_t r;
        message_t *m = (message_t*) c->send_queue->data;

        g_assert(m);
        l = m->length+sizeof(message_t);

        if ((r = write(c->sock, ((guint8*) m) + c->send_index, l-c->send_index)) <= 0) {
            c->dead = TRUE;
            return -1;
        }

        c->send_index += r;

        if (c->send_index >= l) {
            c->send_queue = g_slist_remove(c->send_queue, m);
            c->send_index = 0;

            message_deref(m);
        }

        if (!c->send_queue)
            FD_CLR(c->sock, &listen_wfds);
    }
    return 0;
}

message_t* message_new(message_code_t c, guint8* d, guint s) {
    guchar *p;
    message_t *m;

    if (!d)
        s = 0;

    m = (message_t*) (p = g_new(guint8, sizeof(message_t) + s));

    if (d)
        memcpy(p + sizeof(message_t), d, s);

    m->code = c;
    m->pid = getpid();
    m->length = s;
    m->ref = 1;

    return m;
}

message_t* client_send_enqueue(client_t *c, message_t*m) {
    g_assert(c && m);
    m->ref++;
    c->send_queue = g_slist_append(c->send_queue, m);
    FD_SET(c->sock, &listen_wfds);
    return m;
}


message_t* broadcast_send_enqueue(message_t *m) {
    GSList *l;
    for (l = clients; l; l = l->next)
        client_send_enqueue((client_t*) l->data, m);
    return m;
}

void add_packet(ipq_packet_msg_t *m) {
    ipq_packet_msg_t *n;
    guint l = sizeof(ipq_packet_msg_t)+ m->data_len;
     n = (ipq_packet_msg_t*) g_malloc(l);
    memcpy(n, m, l);
    packets = g_slist_prepend(packets, n);
}

void release_packet(ipq_packet_msg_t *m) {
    packets = g_slist_remove(packets, m);
    g_free(m);
}

ipq_packet_msg_t* find_packet(unsigned long id) {
    GSList *l;
    for (l = packets; l; l = l->next) {
        ipq_packet_msg_t *m = (ipq_packet_msg_t*) l->data;
        if (m->packet_id == id)
            return m;
    }

    return NULL;
}

client_t* find_client(int fd) {
    GSList *l;

    for (l = clients; l; l = l->next) {
        client_t *c = (client_t*) l->data;
        g_assert(c);
        if (c->sock == fd)
            return c;
    }

    return NULL;
}

int new_client(int fd) {
    client_t *c;

    if (find_client(fd))
        return -1;

    g_message("New client connected");
    
    c = g_new(client_t, 1);
    c->sock = fd;
    c->pid = (pid_t) -1;
    c->send_queue = NULL;
    c->send_index = 0;
    c->recv_buf = g_new(guint8, BUFSIZE);
    c->recv_index = 0;
    c->recv_msg = NULL;
    c->dead = FALSE;

    clients = g_slist_prepend(clients, c);

    FD_SET(c->sock, &listen_rfds);
    return 0;
}

int remove_client(int fd) {
    client_t *c;
    
    if (!(c = find_client(fd)))
        return -1;

    c->dead = TRUE;
    return 0;
}

void free_client(client_t *c) {
    GSList *l;

    g_message("Client disconnected");

    FD_CLR(c->sock, &listen_rfds);
    close(c->sock);
    for (l = c->send_queue; l; l = l->next)
        g_free(l->data);
    g_slist_free(c->send_queue);
    g_free(c->recv_buf);
    if (c->recv_msg)
        message_deref(c->recv_msg);
    g_free(c);
}


void kill_clients(gboolean all) {
    if (all) {
        while (clients) {
            client_t *c = (client_t*) clients->data;
            clients = g_slist_remove(clients, c);
            free_client(c);
        }
    } else {
        GSList *prev = NULL, *l;

        for (l = clients; l;) {
            client_t *c = (client_t*) clients->data;
            if (c->dead) {
                GSList *n;
                if (prev)
                    prev->next = l->next;
                else
                    clients = l->next;

                n = l;
                l = l->next;
                free_client((client_t*) n->data);
                g_slist_free_1(n);
            } else {
                prev = l;
                l = l->next;
            }
        }
    }
}

int client_recv_work(client_t *c) {
    ssize_t r;
    size_t l;

    if (!c->recv_buf) {
        c->recv_buf = g_new(guchar, BUFSIZE);
        c->recv_index = 0;
    }
    
    if (c->recv_index >= sizeof(message_t)) {
        l = ((message_t*) c->recv_buf)->length + sizeof(message_t);

        if (l > BUFSIZE) {
            c->dead = TRUE;
            return -1;
        }
    } else
        l = sizeof(message_t);

    if ((r = read(c->sock, c->recv_buf + c->recv_index, l - c->recv_index)) <= 0) {
        c->dead = TRUE;
        return -1;
    }

    c->recv_index += r;

    if (c->recv_index >= sizeof(message_t)) {
        if (c->recv_index >= ((message_t*) c->recv_buf)->length + sizeof(message_t)) {
            c->recv_msg = (message_t*) c->recv_buf;
            c->recv_buf = NULL;
            c->recv_index = 0;
        }
    }
    return 0;
}

message_t *client_recv_pop(client_t *c) {
    if (c->recv_msg) {
        message_t *m = c->recv_msg;
        c->recv_msg = NULL;
        return m;
    }
    return NULL;
}


guint16 csum(guint8 *p, guint l) {
    guint32 sum = 0;
    guint i;
    for (i = 0; i < l; i += 2) 
        sum += (p[i] << 8) + p[i+1];

    sum = (sum & 0xffff) + (sum >> 16);
    sum += (sum >> 16);
    if (sum == 0xffff) sum = 0;

    return ~sum;
}


int reply_icmp(int sock, ipq_packet_msg_t *m, int code) {
    static guint8 buf[BUFSIZE];
    struct iphdr* ip = (struct iphdr*) m->payload;
    struct icmphdr *icmp = (struct icmphdr*) buf;
    guint l, tl;
    struct sockaddr_in sa;
    
    icmp->type = 3;
    icmp->code = code;
    icmp->checksum = 0;
    icmp->un.frag.__unused = 0;
    icmp->un.frag.mtu = 0;

    l = ip->ihl*4 + 8;
    tl = l + sizeof(struct icmphdr);

    memcpy(&buf[sizeof(struct icmphdr)], m->payload, l);

    icmp->checksum = htons(csum((guint8*) icmp, tl));

    sa.sin_family = AF_INET;
    sa.sin_port = 0;
    sa.sin_addr.s_addr = ip->saddr;

    if (sendto(sock, icmp, tl, 0, (struct sockaddr*) &sa, sizeof(sa)) < 0) {
        perror("sendto()");
        return -1;
    }
    
    return 0;
}

int verdict(struct ipq_handle *ipq, int sock, ipq_packet_msg_t *m, int resp) {
    int nf = NF_DROP, icmp = 0;

    if (resp < 0) resp = 0;
    if (resp > 17) resp = 17;
    
    if (resp == 0)
        nf = NF_DROP;
    else if (resp == 1)
        nf = NF_ACCEPT;
    else {
        nf = NF_DROP;
        icmp = 1;
    }
    
    if (ipq_set_verdict(ipq, m->packet_id, nf, 0, NULL) < 0) {
        ipq_perror("ipq_set_verdict");
        return -1;
    }
    
    if (icmp)
        return reply_icmp(sock, m, resp-2);

    return 0;
}

int icmp_socket() {
    int sock, enable = 1;
    
    if ((sock = socket(PF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
        perror("socket(PF_INET, SOCK_RAW, IPPROTO_ICMP)");
        return -1;
        
    }
    
    if (setsockopt(sock, SOL_SOCKET, SO_BROADCAST, (char *)&enable, sizeof(enable)) < 0) {
        perror("setsockopt(.., SOL_SOCKET, SO_BROADCAST)");
        return -1;
    }

    return sock;
}

int main(int argc, char *argv[]) {
    int status;
    unsigned char buf[BUFSIZE];
    struct ipq_handle *ipq = NULL;
    int r = 1;
    int sock = -1, icmp_sock = -1;
    struct group *gr;

    setuid(geteuid());
    
    signal(SIGINT, sigfunc);
    siginterrupt(SIGINT, TRUE);
    signal(SIGQUIT, sigfunc);
    siginterrupt(SIGQUIT, TRUE);
    signal(SIGPIPE, SIG_IGN);
    siginterrupt(SIGPIPE, FALSE);

    
    if (!(ipq = ipq_create_handle(0, PF_INET))) { 
        ipq_perror("ipq_create_handle"); 
        goto finish; 
    } 
    
    if (!(gr = getgrnam("fieryfilter"))) {
        fprintf(stderr, "Could not find group fieryfilter\n");
        goto finish;
    }
    
    if (setgid(gr->gr_gid) != 0) {
        perror("setgid()"); 
        goto finish; 
    } 
    
    if ((sock = unix_socket(SOCKET_PATH)) < 0)
        goto finish;

    if ((icmp_sock = icmp_socket()) < 0)
        goto finish;
    
    
    if ((status = ipq_set_mode(ipq, IPQ_COPY_PACKET, BUFSIZE)) < 0) { 
        ipq_perror("ipq_set_mode"); 
        goto finish; 
    } 

    FD_ZERO(&listen_wfds);
    FD_ZERO(&listen_rfds);

    FD_SET(ipq->fd, &listen_rfds);
    FD_SET(sock, &listen_rfds);

    fprintf(stderr, "ffd started successfully.\n");
    
    for (;;) {
        fd_set rfds = listen_rfds, wfds = listen_wfds;
        
        if (select(FD_SETSIZE, &rfds, &wfds, NULL, NULL) < 0) {
            if (errno == EINTR) {
                switch (signum) {
                    case SIGINT:
                    case SIGQUIT:
                        fprintf(stderr, "Recieved %s, exiting cleanly.\n", strsignal(signum));
                        r = 0;
                        goto finish;
                }
            } else {
                perror("select()");
                goto finish;
            }
        } else {
            GSList *l;

            if (FD_ISSET(sock, &rfds)) {
                int fd;
                if ((fd = accept(sock, NULL, NULL)) < 0) {
                    perror("accept()");
                    goto finish;
                }

                new_client(fd);
            }

            for (l = clients; l; l = l->next) {
                client_t* c = (client_t*) l->data;
                message_t *m;
                g_assert(c);

                if (!c->dead && FD_ISSET(c->sock, &wfds))
                    client_send_work(c);

                if (!c->dead && FD_ISSET(c->sock, &rfds))
                    client_recv_work(c);

                if (!c->dead && (m = client_recv_pop(c))) {
                    switch (m->code) {
                        case MSG_HELLO:
                            fprintf(stderr, "Client %u connected.\n", m->pid);
                            message_deref(client_send_enqueue(c, message_new(MSG_HELLO, NULL, 0)));
                            break;
                            
                        case MSG_GOODBYE:
                            fprintf(stderr, "Client %u disconnected.\n", m->pid);
                            message_deref(client_send_enqueue(c, message_new(MSG_GOODBYE, NULL, 0)));
                            break;

                        case MSG_VERDICT: {
                            ipq_packet_msg_t *ipqm;
                            unsigned long packet_id =  *((unsigned long*) (m+1));
                            
                            if (m->length != sizeof(unsigned long)+sizeof(guint32)) {
                                fprintf(stderr, "Recieved verdict with bogus size, killing client.\n");
                                c->dead = TRUE;
                                break;
                            }

                            fprintf(stderr, "Recieved verdict message from %u for %lu.\n", m->pid, packet_id);
                            
                            if ((ipqm = find_packet(packet_id))) {
                                if (verdict(ipq, icmp_sock, ipqm, *((guint32*) ((guint8*) m+sizeof(message_t) + sizeof(unsigned long)))) < 0) {
                                    release_packet(ipqm);
                                    fprintf(stderr, "Could not set verdict.\n");
                                    message_deref(m);
                                    goto finish;
                                }
                                release_packet(ipqm);
                            } else {
                                fprintf(stderr, "Recieved verdict for unknown packet id, killing client.\n");
                                c->dead = TRUE;
                            }
                            
                            break;
                        }
                        
                        default:

                            fprintf(stderr, "Recieved bogus message from client, killing client\n");
                            c->dead = TRUE;
                            break;
                    }

                    message_deref(m);
                }

            }

            if (FD_ISSET(ipq->fd, &rfds)) {
                if ((status = ipq_read(ipq, buf, BUFSIZE, 0)) < 0) {
                    ipq_perror("ipq_read");
                    goto finish;
                }


                switch (ipq_message_type(buf)) {
                    case NLMSG_ERROR:
                        fprintf(stderr, "Received error message: %s\n", strerror(ipq_get_msgerr(buf)));
                        goto finish;
                        
                    case IPQM_PACKET: {
                        ipq_packet_msg_t *m = ipq_get_packet(buf);
                        struct iphdr *ip = (struct iphdr*) m->payload;
                        struct tcphdr *tcp = (struct tcphdr*) (((guint8*)m->payload)+ip->ihl*4);
                        guint32 a, b;
                        gchar at[20], bt[256];
                        guint16 po;

                        if (!m->timestamp_sec)
                            m->timestamp_sec = time(0);

                        
                        a = ip->saddr;
                        b = ip->daddr;
                        po = ntohs(tcp->dest);
                            
                        strcpy(at, inet_ntoa(*((struct in_addr*) &a)));
                        strcpy(bt, inet_ntoa(*((struct in_addr*) &b)));
                        
                        fprintf(stderr, "(%lu) Recieved packet (%s->%s) (%s->%s:%u) [%lu] len=%i\n", m->packet_id, m->indev_name, m->outdev_name, at, bt, po, m->timestamp_sec, m->data_len);

                        add_packet(m);

                        if (clients)
                            message_deref(broadcast_send_enqueue(message_new(MSG_PACKET, (guint8*) m, sizeof(ipq_packet_msg_t)+m->data_len)));
                        else {
                            fprintf(stderr, "(%lu) No clients listening, sending verdict.\n", m->packet_id);
                            if (verdict(ipq, icmp_sock, m, 15) < 0) {
                                fprintf(stderr, "Sending verdict failed.\n");
                                goto finish;
                            }
                        }
                        
                        break;
                    }
                        
                    default:
                        fprintf(stderr, "Unknown message type!\n");
                        goto finish;
                }
                
            }
        }
        
        kill_clients(FALSE);
    }
    
    r = 0;
    
finish:
    
    fprintf(stderr, "ffd exiting.\n");

    kill_clients(TRUE);

    if (sock != -1) {
        close(sock);
        remove(SOCKET_PATH);
    }

    if (icmp_sock != -1)
        close(icmp_sock);
    
    if (ipq)
        ipq_destroy_handle(ipq);
    
    return r;
}
